use std::fmt::{self, Formatter};
use std::time::{SystemTime, UNIX_EPOCH};

use aws_config::BehaviorVersion;
use aws_smithy_types::date_time::{DateTime, Format};
use chacha20poly1305::XNonce;
use serde::{
    Serialize,
    Deserialize,
    Serializer,
    Deserializer,
};
use serde::de::{self, Visitor};
use sqlx::{
    FromRow,
    Sqlite,
    Transaction,
    types::Uuid,
};

use super::{Crypto, PersistentCredential};

use crate::errors::*;


#[derive(Debug, Clone, FromRow)]
pub struct AwsRow {
    #[allow(dead_code)]
    id: Uuid,
    access_key_id: String,
    secret_key_enc: Vec<u8>,
    nonce: Vec<u8>,
}


#[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct AwsBaseCredential {
    #[serde(default = "default_credentials_version")]
    pub version: usize,
    pub access_key_id: String,
    pub secret_access_key: String,
}


impl AwsBaseCredential {
    pub fn new(access_key_id: String, secret_access_key: String) -> Self {
        Self {version: 1, access_key_id, secret_access_key}
    }
}

impl PersistentCredential for AwsBaseCredential {
    type Row = AwsRow;

    fn type_name() -> &'static str { "aws" }

    fn from_row(row: AwsRow, crypto: &Crypto) -> Result<Self, LoadCredentialsError> {
        let nonce = XNonce::clone_from_slice(&row.nonce);
        let secret_key_bytes = crypto.decrypt(&nonce, &row.secret_key_enc)?;
        let secret_key = String::from_utf8(secret_key_bytes)
            .map_err(|_| LoadCredentialsError::InvalidData)?;

        Ok(Self::new(row.access_key_id, secret_key))
    }

    async fn save_details(&self, id: &Uuid, crypto: &Crypto, txn: &mut Transaction<'_, Sqlite>) -> Result<(), SaveCredentialsError> {
        let (nonce, ciphertext) = crypto.encrypt(self.secret_access_key.as_bytes())?;
        let nonce_bytes = &nonce.as_slice();

        sqlx::query!(
            "INSERT OR REPLACE INTO aws_credentials (
                id,
                access_key_id,
                secret_key_enc,
                nonce
            ) 
            VALUES (?, ?, ?, ?);",
            id, self.access_key_id, ciphertext, nonce_bytes,
        ).execute(&mut **txn).await?;

        Ok(())
    }

    // async fn save(&self, record: CredentialRecord, &Crypto, pool: &SqlitePool) -> Result<(), CredentialRecordsError> {
    //     let (nonce, ciphertext) = crypto.encrypt(self.secret_access_key.as_bytes())?;
    //     let nonce_bytes = &nonce.as_slice();

    //     let res = sqlx::query!(
    //         "INSERT INTO credentials (id, name, type, created_at)
    //         VALUES (?, ?, 'aws', strftime('%s'))
    //         ON CONFLICT(id) DO UPDATE SET
    //             name = excluded.name,
    //             type = excluded.type,
    //             created_at = excluded.created_at;
            
    //         INSERT OR REPLACE INTO aws_credentials (
    //             id,
    //             access_key_id,
    //             secret_key_enc,
    //             nonce
    //         ) 
    //         VALUES (?, ?, ?, ?);",
    //         id,
    //         name,
    //         id, // for the second query
    //         self.access_key_id,
    //         ciphertext,
    //         nonce_bytes,
    //     ).execute(pool).await;

    //     match res {
    //         Err(SqlxError::Database(e)) if e.code().as_deref() == Some("2067") => Err(CredentialRecordsError::Duplicate),
    //         Err(e) => Err(SaveCredentialsError::DbError(e)),
    //         Ok(_) => Ok(())
    //     }
    // }

    // async fn load(name: &str, crypto: &Crypto, pool: &SqlitePool) -> Result<Self, LoadCredentialsError> {
    //     let record: AwsRecord = sqlx::query_as(
    //         "SELECT c.id, c.name, c.is_default, a.access_key_id, a.secret_key_enc, a.nonce
    //         FROM credentials c JOIN aws_credentials a ON a.id = c.id
    //         WHERE c.name = ?"
    //     ).bind(name)
    //     .fetch_optional(pool)
    //     .await?
    //     .ok_or(LoadCredentialsError::NoCredentials)?;

    //     let key = record.decrypt_key(crypto)?;
    //     let credential = AwsBaseCredential::new(record.access_key_id, key);
    //     Ok(credential)
    // }

    // async fn load_default(crypto: &Crypto, pool: &SqlitePool) -> Result<Self, LoadCredentialsError> {
    //     let record: AwsRecord = sqlx::query_as(
    //         "SELECT c.id, c.name, c.is_default, a.access_key_id, a.secret_key_enc, a.nonce
    //         FROM credentials c JOIN aws_credentials a ON a.id = c.id
    //         WHERE c.type = 'aws' AND c.is_default = 1"
    //     ).fetch_optional(pool)
    //     .await?
    //     .ok_or(LoadCredentialsError::NoCredentials)?;

    //     let key = record.decrypt_key(crypto)?;
    //     let credential = AwsBaseCredential::new(record.access_key_id, key);
    //     Ok(credential)
    // }

    // async fn list(crypto: &Crypto, pool: &SqlitePool) -> Result<Vec<SaveCredential>, LoadCredentialsError> {
    //     let mut rows = sqlx::query_as::<_, AwsRecord>(
    //         "SELECT c.id, c.name, c.is_default, a.access_key_id, a.secret_key_enc, a.nonce
    //         FROM credentials c JOIN aws_credentials a ON a.id = c.id"
    //     ).fetch(pool);

    //     let mut creds = Vec::new();

    //     while let Some(record) = rows.try_next().await? {
    //         let key = record.decrypt_key(crypto)?;
    //         let aws = AwsBaseCredential::new(record.access_key_id, key);

    //         let cred = SaveCredential {
    //             id: record.id,
    //             name: record.name,
    //             is_default: record.is_default,
    //             credential: Credential::AwsBase(aws),
    //         };
    //         creds.push(cred);
    //     }

    //     Ok(creds)
    // }
}


#[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct AwsSessionCredential {
    #[serde(default = "default_credentials_version")]
    pub version: usize,
    pub access_key_id: String,
    pub secret_access_key: String,
    pub session_token: String,
    #[serde(serialize_with = "serialize_expiration")]
    #[serde(deserialize_with = "deserialize_expiration")]
    pub expiration: DateTime,
}

impl AwsSessionCredential {
    pub async fn from_base(base: &AwsBaseCredential) -> Result<Self, GetSessionError> {
        let req_creds = aws_sdk_sts::config::Credentials::new(
            &base.access_key_id,
            &base.secret_access_key,
            None, // token
            None, //expiration
            "Creddy", // "provider name" apparently
        );
        let config = aws_config::defaults(BehaviorVersion::latest())
            .credentials_provider(req_creds)
            .load()
            .await;

        let client = aws_sdk_sts::Client::new(&config);
        let resp = client.get_session_token()
            .duration_seconds(43_200)
            .send()
            .await?;

        let aws_session = resp.credentials.ok_or(GetSessionError::EmptyResponse)?;

        let session_creds = AwsSessionCredential {
            version: 1,
            access_key_id: aws_session.access_key_id,
            secret_access_key: aws_session.secret_access_key,
            session_token: aws_session.session_token,
            expiration: aws_session.expiration,
        };

        #[cfg(debug_assertions)]
        println!("Got new session:\n{}", serde_json::to_string(&session_creds).unwrap());

        Ok(session_creds)
    }

    pub fn is_expired(&self) -> bool {
        let current_ts = SystemTime::now()
            .duration_since(UNIX_EPOCH)
            .unwrap() // doesn't panic because UNIX_EPOCH won't be later than now()
            .as_secs();

        let expire_ts = self.expiration.secs();
        let remaining = expire_ts - (current_ts as i64);
        remaining < 60
    }
}


fn default_credentials_version() -> usize { 1 }


struct DateTimeVisitor;

impl<'de> Visitor<'de> for DateTimeVisitor {
    type Value = DateTime;

    fn expecting(&self, formatter: &mut Formatter) -> fmt::Result {
        write!(formatter, "an RFC 3339 UTC string, e.g. \"2014-01-05T10:17:34Z\"")
    }

    fn visit_str<E: de::Error>(self, v: &str) -> Result<DateTime, E> {
        DateTime::from_str(v, Format::DateTime)
            .map_err(|_| E::custom(format!("Invalid date/time: {v}")))
    }
}


fn deserialize_expiration<'de, D>(deserializer: D) -> Result<DateTime, D::Error>
where D: Deserializer<'de>
{
    deserializer.deserialize_str(DateTimeVisitor)
}

fn serialize_expiration<S>(exp: &DateTime, serializer: S) -> Result<S::Ok, S::Error>
where S: Serializer
{
    // this only fails if the d/t is out of range, which it can't be for this format
    let time_str = exp.fmt(Format::DateTime).unwrap();
    serializer.serialize_str(&time_str)
}


#[cfg(test)]
mod tests {
    use super::*;
    use sqlx::SqlitePool;


    fn test_creds() -> AwsBaseCredential {
        AwsBaseCredential::new(
            "AKIAIOSFODNN7EXAMPLE".into(),
            "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY".into(),
        )
    }

    fn test_creds_2() -> AwsBaseCredential {
        AwsBaseCredential::new(
            "AKIAIOSFODNN7EXAMPL2".into(),
            "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKE2".into(),
        )
    }

    fn test_uuid() -> Uuid {
        Uuid::try_parse("00000000-0000-0000-0000-000000000000").unwrap()
    }

    fn test_uuid_2() -> Uuid {
        Uuid::try_parse("ffffffff-ffff-ffff-ffff-ffffffffffff").unwrap()
    }

    fn test_uuid_random() -> Uuid {
        let bytes = Crypto::salt();
        Uuid::from_slice(&bytes[..16]).unwrap()
    }


    #[sqlx::test]
    async fn test_save(pool: SqlitePool) {
        let crypt = Crypto::random();
        let mut txn = pool.begin().await.unwrap();
        test_creds().save_details(&test_uuid_random(), &crypt, &mut txn).await
            .expect("Failed to save AWS credentials");
    }


    #[sqlx::test(fixtures("aws_credentials"))]
    async fn test_overwrite(pool: SqlitePool) {
        let crypt = Crypto::fixed();

        let creds = test_creds_2();
        // overwite original creds with different test data
        let mut txn = pool.begin().await.unwrap();
        creds.save_details(&test_uuid(), &crypt, &mut txn).await
            .expect("Failed to update AWS credentials");

        // make sure update went through
        let loaded = AwsBaseCredential::load(&test_uuid(), &crypt, &pool).await.unwrap();
        assert_eq!(creds, loaded);
    }


    #[sqlx::test(fixtures("aws_credentials"))]
    async fn test_duplicate_name(pool: SqlitePool) {
        let crypt = Crypto::random();

        let id = test_uuid_random();
        let mut txn = pool.begin().await.unwrap();
        let resp = test_creds().save_details(&id, &crypt, &mut txn).await;

        if !matches!(resp, Err(SaveCredentialsError::Duplicate)) {
            panic!("Attempt to create duplicate entry returned {resp:?}")
        }
    }


    #[sqlx::test(fixtures("aws_credentials"))]
    async fn test_load(pool: SqlitePool) {
        let crypt = Crypto::fixed();
        let loaded = AwsBaseCredential::load(&test_uuid(), &crypt, &pool).await.unwrap();
        assert_eq!(test_creds(), loaded);
    }


    #[sqlx::test(fixtures("aws_credentials"))]
    async fn test_load_by_name(pool: SqlitePool) {
        let crypt = Crypto::fixed();
        let loaded = AwsBaseCredential::load_by_name("test", &crypt, &pool).await.unwrap();
        assert_eq!(test_creds(), loaded);
    }


    #[sqlx::test]
    async fn test_save_load(pool: SqlitePool) {
        let crypt = Crypto::random();
        let creds = test_creds();
        let id = test_uuid_random();

        let mut txn = pool.begin().await.unwrap();
        creds.save_details(&id, &crypt, &mut txn).await.unwrap();
        let loaded = AwsBaseCredential::load(&id, &crypt, &pool).await.unwrap();

        assert_eq!(creds, loaded);
    }

    // #[sqlx::test(fixtures("aws_credentials"))]
    // async fn test_list(pool: SqlitePool) {
    //     let crypt = Crypto::fixed();
    //     let list = AwsBaseCredential::list(&crypt, &pool).await
    //         .expect("Failed to list AWS credentials");

    //     let first = SaveCredential {
    //         id: test_uuid(),
    //         name: "test".into(),
    //         credential: Credential::AwsBase(test_creds()),
    //     };
    //     assert_eq!(&first, &list[0]);

    //     let second = SaveCredential {
    //         id: test_uuid_2(),
    //         name: "test2".into(),
    //         credential: Credential::AwsBase(test_creds_2()),
    //     };
    //     assert_eq!(&second, &list[1]);
    // }

    // #[sqlx::test(fixtures("aws_credentials"))]
    // async fn test_rekey(pool: SqlitePool) {
    //     let old_crypt = Crypto::fixed();
    //     let orig = AwsBaseCredential::list(&old_crypt, &pool).await.unwrap();

    //     let new_crypt = Crypto::random();
    //     AwsBaseCredential::rekey(&old_crypt, &new_crypt, &pool).await
    //         .expect("Failed to re-key AWS credentials");
        
    //     let rekeyed = AwsBaseCredential::list(&new_crypt, &pool).await.unwrap();
    //     for (before, after) in orig.iter().zip(rekeyed.iter()) {
    //         assert_eq!(before, after);
    //     }
    // }
}
